/***
* Code.prg
*
* Compile with /a /m /n /w
*
* Run code.exe passing the file name on the command line
*/

#include "fileio.ch"

#define CR    chr(13)
#define LF    chr(10)
#define CRLF  CR + LF
#define FTELL(handle)  fseek(handle, 0, FS_RELATIVE)

#include "inkey.ch"

#command DEFAULT <p> TO <val> [,<pn> TO <valn>] ;
         => ;
         <p> = iif(<p> = NIL, <val>, <p>) ;
         [;<pn> = iif(<pn> = NIL, <valn>, <pn>)]


STATIC nHandle, cLine,nFileSize


FUNCTION test(fname)

LOCAL to, tbco, tbco1

    CLEAR SCREEN
    @ 0, 0 TO 24, 79

    lFileInit(fname)
    to := tbrowsenew(1, 1, 23, 78)
    tbco := tbColumnNew(, {|| cCurLine() })
    tbco1 := tbcolumnnew(, {|| substr(cCurLine(), 79) })

    tbco:width := 78
    tbco1:width := 78
    to:addcolumn(tbco)
    to:addcolumn(tbco1)
    to:goBottomBlock := {|| GetLast() }
    to:goTopBlock    := {|| GetFirst() }
    to:skipBlock := {|n| skipper(n) }
    tbrowse(to)

RETURN NIL


FUNCTION skipper(n)

LOCAL nSkipped := 0

    IF n > 0
        DO WHILE nSkipped != n .AND. lGoNextLn()
            nSkipped++
        ENDDO
    ELSE
        DO WHILE nSkipped != n .AND. lGoPrevLn()
            nSkipped--
        ENDDO
    ENDIF

RETURN nSkipped

#define MAX_LINE_LEN 256

FUNCTION lFileInit(nFname)

    IF (nHandle := fopen(nFname)) > 0
        fseek(nHandle, 0, FS_END)
        nFileSize := FTELL(nHandle)
        GetFirst()
    ENDIF

RETURN nHandle != 0

FUNCTION GetLast

    fseek(nHandle, 0, FS_END)
    lGoPrevLn()

RETURN NIL

FUNCTION GetFirst

    fseek(nHandle, 0, FS_SET)
    lfreadln(nHandle, @cline)
    fseek(nHandle, 0, FS_SET)

RETURN NIL


FUNCTION lGoPrevLn

LOCAL nOrigPos := FTELL(nHandle), nMaxRead, nNewPos, ;
      lMoved, cBuff, nWhereCrLf, nPrev

    IF nOrigPos = 0
        lMoved := .F.
    ELSE
        lMoved := .T.
        IF nOrigPos != nFileSize
            //  Skip over preceeding CR / LF
            fseek(nHandle, -2, FS_RELATIVE)
        ENDIF
        nMaxRead := MIN(MAX_LINE_LEN, FTELL(nHandle))

        cBuff := space(nMaxRead)
        nNewPos := fseek(nHandle, -nMaxRead, FS_RELATIVE)
        fread(nHandle, @cBuff, nMaxRead)
        nWhereCrLf := rat(CRLF, cBuff)
        IF nWhereCrLf = 0
            nPrev := nNewPos
            cLine = cBuff
        ELSE
            nPrev := nNewPos + nWhereCrLf + 1
            cline := substr(cBuff, nWhereCrLf + 2)
        ENDIF

        fseek(nHandle, nPrev, FS_SET)
    ENDIF

RETURN lMoved


FUNCTION lGoNextLn()

LOCAL nSavePos := FTELL(nHandle), ;
      cBuff := "", lMoved, nNewPos

    fseek(nHandle, len(cLine) + 2, FS_RELATIVE)
    nNewPos := FTELL(nHandle)
    IF lfreadLn(nHandle, @cBuff)
        lMoved := .T.
        cLine := cBuff
        fseek(nHandle, nNewPos, FS_SET)
    ELSE
        lMoved := .F.
        fseek(nHandle, nSavePos, FS_SET)
    ENDIF

RETURN lMoved


FUNCTION cCurLine

RETURN cLine

FUNCTION lfreadln(nHandle, cBuffer)

LOCAL nEol, nRead, nSaveFpos

    cBuffer = space(MAX_LINE_LEN)

    * First save current file pointer
    nSaveFpos = FTELL(nHandle)
    nRead = fread(nHandle, @cBuffer, MAX_LINE_LEN)

    IF (nEol := at(CRLF, substr(cBuffer, 1, nRead))) = 0
        * Line overflow or eof
        * Cline has the line we need
    ELSE
        * Copy up to eol
        cBuffer = substr(cBuffer, 1, nEol - 1)
        * Position file pointer to next line
        fseek(nHandle, nSaveFpos + nEol + 1, FS_SET)
    ENDIF

RETURN nRead != 0


    * Pass tbrowse() a tbrowse object, and up to three code blocks.
    * The first is the code block to check for keys before checking the
    * standard keys, the second indicates how long the user stays in
    * the browse, the third the code block to check keys after the
    * standard keys have been checked.

    FUNCTION tbrowse(toThis, bBefore, ;
                     bWhileCond, bAfter)

    LOCAL nKey, lExitRequested := .F.

        DEFAULT bBefore TO {|nKey| ;
                             iif(nKey = K_ENTER .OR. ;
                                 nKey = K_ESC, ;
                                 lExitRequested := .T., ;
                                 .F.) ;
                           }

        DEFAULT bWhileCond TO {|| !lExitRequested}

        DEFAULT bAfter TO {|| .F.}

        DO WHILE eval(bWhileCond, toThis)
            DO WHILE nextkey() = 0 .AND. ;
                     !toThis:stabilize()
            ENDDO

            nKey = inkey(0)
            IF eval(bBefore, nKey, toThis)
                // Processed key as before key
            ELSEIF stdmeth(nKey, toThis)
                // Processed key as standard key
            ELSEIF toThis:stable .AND. ;
                   eval(bAfter, nKey, toThis)
                // Processed key as after key
            ENDIF
        ENDDO

    RETURN NIL

    * Attempt to process nKey as a standard method. Returns .T. if
    * processed, .F. otherwise.

    FUNCTION stdmeth(nKey, toThis)

    LOCAL lKeyHandled := .T.

        DO CASE
            CASE nKey = K_DOWN;       toThis:down()
            CASE nKey = K_UP;         toThis:up()
            CASE nKey = K_PGDN;       toThis:pagedown()
            CASE nKey = K_PGUP;       toThis:pageup()
            CASE nKey = K_CTRL_PGUP;  toThis:gotop()
            CASE nKey = K_CTRL_PGDN;  toThis:gobottom()
            CASE nKey = K_RIGHT;      toThis:right()
            CASE nKey = K_LEFT;       toThis:left()
            CASE nKey = K_HOME;       toThis:home()
            CASE nKey = K_END;        toThis:end()
            CASE nKey = K_CTRL_LEFT;  toThis:panleft()
            CASE nKey = K_CTRL_RIGHT; toThis:panright()
            CASE nKey = K_CTRL_HOME;  toThis:panhome()
            CASE nKey = K_CTRL_END;   toThis:panend()
            OTHERWISE;                lKeyHandled := .F.
        ENDCASE

    RETURN lKeyHandled
